(*
<one line to give the program's name and a brief idea of what it does.>
Copyright (C) 2012 Hendrik Boom

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)
open GL
open Glu
open Glut
open Printf
open Unix

let error() =
  let err = glGetError() in
  if err <> GL_NO_ERROR
  then printf "%s\n%!" (gluErrorString err)
;;

let hills = true
let worldradius = 43
let worldwidth = (* 85 *) (worldradius * 2 - 1)
let sealevel = 0.0
let tall = 0.25
let b_down = ref false
let anglex = ref 0.0
let angley = ref 0.0
let xold = ref 0
let yold = ref 0
let locx = ref ((float_of_int worldwidth) /. 2.0)
let locy = ref ((float_of_int worldwidth) /. 2.0)
let lookdown = ref false
let windowwidth = ref 640
let windowheight = ref 400
;;
(* This should work:
    Random.self_init;
  But in the ocaml avaliable here, it doesn't.  Every time the same world is generated.  So  I had to change it.
*)
let date = (Unix.gmtime (Unix.time())) in
    Random.init (date.tm_sec + 60 * (date.tm_mon + 60 * (date.tm_hour + 24 * (date.tm_mday + 31 * date.tm_mon))));
;;
let corners = Array.make_matrix (worldwidth+1)  (worldwidth+1) 0
let corneraltitude =Array.make_matrix (worldwidth+1)  (worldwidth+1) sealevel 
;;

let round x = int_of_float (floor (x +. 0.5))
;;

(* These functions define the relationships between (i, j) coordinates and
      (x, z) coordinates

   Terminology: north is in the direction of decdeasing z
                south is the directino of increasing z
                east is the direction of increasing x
                west is the direction of decreasing x
*)

(* Most of these functions ignore one or other of their arguments.
   But both arguments are provided anyway because it's clearer that way.
   Note that these floating point 
*)

let x_of_tilecentre i j = (* UNUSED *)
 (float_of_int i)
let z_of_tilecentre i j = (* UNUSED *)
 (float_of_int j)

let x_of_corner i j = (* UNUSED *)
  (float_of_int i) -. 0.5
let z_of_corner i j = (* UNUSED *) 
  (float_of_int j) -. 0.5

(* Find the nearest vertex to (x, z) *)
let i_of_corner x z = round (x +. 0.5)
let j_of_corner x z = round (z +. 0.5)

(* Find the tile containing (x, z) *)
let i_of_tile x z = round x
let j_of_tile x z = round z
;;

let sqinterp tx tz nw ne se sw =
  let mid = (nw +. ne +. se +. sw ) /. 4.0 in 
(*  printf "m%f\n" mid; *)
  if tx < tz
    then if tx+.tz < 1.0
      then (* west triangle *)
           (  nw *.(1.0-.tx-.tz)
           +. sw *. (tz-.tx)
           +. mid *. tx *. 2.0
           )
      else (* south triangle *)
           ( sw *. (tz-.tx)
           +. se *. (tx+.tz-.1.0)
           +. mid *. (1.0 -. tz) *. 2.0
           )
    else (* north triangle *)
      (* tx > tz *)
      if tx+.tz < 1.0
      then (  nw *. (1.0-.tx-.tz)
           +. ne *. (tx-.tz)
           +. mid *. tz *. 2.0
           )
      else ( ne *. (tx-.tz)
           +. se *. (tx+.tz-.1.0)
           +. mid *. (1.0 -. tx) *. 2.0
           )
      
(*
for j = 0 to 6 do
  for i = 0 to 6 do
      printf "%d%d %f " i j (sqinterp ((float_of_int i) /. 6.0) ((float_of_int j) /. 6.0)
          0.0 2.0 0.0 0.0)
  done;
  printf "\n%!"
done
*)
;;   


let newaltitude x z =
  let i = i_of_tile x z in
  let j = j_of_tile x z in
  let tx = x -. x_of_corner i j in
  let tz = z -. z_of_corner i j in
   sqinterp tx tz corneraltitude.(i).(j)
                  corneraltitude.(i+1).(j)
                  corneraltitude.(i+1).(j+1)
                  corneraltitude.(i).(j+1)
;;
let oldaltitude x z = sealevel
;;
let altitude = if hills then newaltitude else oldaltitude
;;
for i = 0 to worldwidth do
  for j = 0 to worldwidth do
    let th = (max (max (worldwidth - i) i)
                  (max (worldwidth - j) j)) - worldradius
    in
    (* th ranges from 0 to worldradius - 1 *)
    corners.(i).(j) <- (if ( Random.int (worldradius -1) < th)
                       then 0
                       else 1)
  done
done;
;;
let rec revise i j =
  if not (i < 0 || i >= worldwidth || j < 0 || j >= worldwidth)
  then
  ( ( if
           ( corners.(i).(j) = 0)
        && (corners.(i).(j+1) = 1)
        && (corners.(i+1).(j+1) = 0)
        && (corners.(i+1).(j) = 1)
      then ( corners.(i).(j) <- 1;
             corners.(i+1).(j+1) <- 1;
             revise (i - 1) (j - 1);
             revise (i + 1) (j + 1);
           )
    );

    ( if
        (    corners.(i).(j) = 1
        && corners.(i).(j+1) = 0
        && corners.(i+1).(j+1) = 1
        && corners.(i+1).(j) = 0
        )
      then ( corners.(i).(j+1) <- 1;
             corners.(i+1).(j) <- 1;
             revise (i - 1) (j + 1);
             revise (i + 1) (j - 1)
           )
    )
  );
;;

for i = 0 to (worldwidth - 1) do
  for j = 0 to (worldwidth - 1) do
    revise i j
  done
done;

if hills
then
 for i = 1 to worldwidth - 1 do
   for j = 1 to worldwidth - 1 do
    if
     corners.(i-1).(j-1) = 1 && corners.(i).(j-1) = 1 && corners.(i+1).(j-1) = 1 &&
     corners.(i-1).(j  ) = 1 && corners.(i).(j  ) = 1 && corners.(i+1).(j  ) = 1 &&
     corners.(i-1).(j+1) = 1 && corners.(i).(j+1) = 1 && corners.(i+1).(j+1) = 1
    then corneraltitude.(i).(j) <- (Random.float 0.25)
   done
 done

;;

let accesscorner = Array.make_matrix (worldwidth+1)  (worldwidth+1) false
let flagcorner = Array.make_matrix (worldwidth+1)  (worldwidth+1) false
let flagfound = Array.make_matrix (worldwidth+1)  (worldwidth+1) false
let nmb_flags_found = ref 0
;;
let rec access i j =
  (
  if i < 0 || i >= worldwidth || j < 0 || j >= worldwidth
  then ()
  else if accesscorner.(i).(j)
  then ()
  else
    (
    accesscorner.(i).(j) <- true;
    if corners.(i).(j) = 1
    then
      if corners.(i+1).(j) = 1
      then access (i+1) j;
      if corners.(i-1).(j) = 1
      then access (i-1) j;
      if corners.(i).(j+1) = 1
      then access i (j+1);
      if corners.(i).(j-1) = 1
      then access i (j-1)
    )
  );

;;

access worldradius worldradius;
(* printf "ink\n%!"; *)

let nmbflags = ref 0 in
while !nmbflags < 8
do
  let i = Random.int worldwidth in
  let j = Random.int worldwidth in
  (if accesscorner.( i).( j) && not flagcorner.(i).(j)
  then (flagcorner.(i).(j) <- true;
        nmbflags := !nmbflags + 1;
        (* corners.(i).(j) <- 2; *)
       )
  )
done;
;;

let istraversible x y = 
  (* This checks whether 'here' is near a corner that contains land.
     It pays no attention to the actual shape of the land in the tile
     containing 'here'
   *)
  let i = i_of_corner x y in
  let j = j_of_corner x y in 
  (* i, j, are integer indices for nearest corner. *)
  if i < 0 || i > worldwidth || j < 0 || j > worldwidth
  then false
  else corners.(i).(j) > 0;
;;

let hasflag x y = 
  let i = i_of_corner x y in
  let j = j_of_corner x y in 
  (* printf "hasflag %d %d\n%!" i j; *)
  if i < 0 || i > worldwidth || j < 0 || j > worldwidth
  then false
  else (if flagcorner.(i).(j)
        then
           (if flagfound.(i).(j)
            then true
            else
              ( flagfound.(i).(j) <- true;
                nmb_flags_found := !nmb_flags_found + 1;
                printf "found flag %d of %d!\n%!" !nmb_flags_found 8;
                true
              )
           )
        else false
    )
;;

let drawmap plot = 
  (for i = 0 to worldwidth do
    (for j = 0 to worldwidth do
       (let c = corners.(i).(j) in
        let f = flagcorner.(i).(j) in
         (if f
          then if flagfound.(i).(j) 
               then plot i j 1.0 1.0 1.0
               else plot i j 1.0 0.0 0.0
          else if c = 1
          then plot i j 0.0 (1.0 -. corneraltitude.(i).(j)) 0.0
          else plot i j 0.0 0.0 1.0
         )
       )
     done
    )
   done;
  plot 1 1 1.0 0.0 1.0;
  )
;;

let drawhmap plot = 
  for j = 0 to 60 do
    for i = 0 to 60 do
       (plot i j 
         (1.0 -. (altitude ((float_of_int worldradius) +. (float_of_int i) /. 10.0)
                           ((float_of_int worldradius) +. (float_of_int j) /. 10.0)
                 )
         )
         0.0 0.0
       )
     done
   done;
;;


(* Random.int 2 for random bits. *)

let tilechoose p q =
  (* TODO: figure out how to get rid of floating point here *)
  glTexCoord2 ((float_of_int p   ) *. 0.0625) ((float_of_int q    ) *. 0.0625)
;;
let tilechoosemid p q =
  (* TODO: figure out how to get rid of floating point here *)
  glTexCoord2 ((float_of_int p +. 0.5)*. 0.0625) ((float_of_int q +. 0.5)*. 0.0625)
;;
let tilechoose0 p q =
  (* TODO: figure out how to get rid of floating point here *)
  glTexCoord2 ((float_of_int p   ) *. 0.0625) ((float_of_int q    ) *. 0.0625)
;;
let tilechoose1 p q =
  glTexCoord2 (float_of_int(p    ) *. 0.0625) (float_of_int (q + 1) *. 0.0625) 
;;
let tilechoose2 p q =
  glTexCoord2 (float_of_int(p + 1) *. 0.0625) (float_of_int (q + 1) *. 0.0625) 
;;
let tilechoose3 p q =
  glTexCoord2 (float_of_int(p + 1) *. 0.0625) (float_of_int (q    ) *. 0.0625) 
;;


let newtileatss (p : int) (q : int) (i : int) (j : int) =
  let x = x_of_corner i j in
  let z = z_of_corner i j in
     (* x, z are coordinates of northwest corner -- the one  with lower z coord *)
  let nw = corneraltitude.(i  ).(j)     in
  let ne = corneraltitude.(i  ).(j+1) in
  let se = corneraltitude.(i+1).(j+1) in
  let sw = corneraltitude.(i+1).(j) in
  let mid = (nw +. ne +. se +. sw) /. 4.0  in
  glEnable GL_TEXTURE_2D;
  glBegin GL_TRIANGLE_FAN;
  tilechoosemid p q;      glVertex3 (x +. 0.5) mid (z +. 0.5);
  tilechoose p     q;     glVertex3 (x)        nw (z);
  tilechoose p     (q+1); glVertex3 (x)        ne (z +. 1.0);
  tilechoose (p+1) (q+1); glVertex3 (x +. 1.0) se (z +. 1.0);
  tilechoose (p+1) q;     glVertex3 (x +. 1.0) sw (z);
  tilechoose p     q;     glVertex3 (x)        nw (z);
  glEnd();
  glDisable GL_TEXTURE_2D;
;;

let oldtileatss (p : int) (q : int) (i : int) (j : int) =
     (* x, z are coordinates of northwest corner -- the one  with lower z coord *)
  let x = x_of_corner i j in
  let z = z_of_corner i j in
  glEnable GL_TEXTURE_2D;
  glBegin GL_QUADS;
  tilechoose p     q;     glVertex3 (x)        sealevel (z);
  tilechoose p     (q+1); glVertex3 (x)        sealevel (z +. 1.0);
  tilechoose (p+1) (q+1); glVertex3 (x +. 1.0) sealevel (z +. 1.0);
  tilechoose (p+1) q;     glVertex3 (x +. 1.0) sealevel (z);
  glEnd();
  glDisable GL_TEXTURE_2D;
;;
let tileatss = if hills then newtileatss else oldtileatss
;;
let flagat x z = ();
(
  (* printf "flagat1\n%!"; *)
 error();
  (* printf "flagat2\n%!"; *)
  let size = 0.1 in
  let axz = altitude x z in 
  glBegin GL_QUADS;
  glColor3 1.0 0.0 0.0;
  glVertex3 (x -. size) (axz) (z);
  glVertex3 (x +. size) (axz) (z);
  glVertex3 (x +. size) (axz +. size *. 2.0) (z);
  glVertex3 (x -. size) (axz +. size *. 2.0) (z);
  glColor3 1.0 1.0 1.0;
  glVertex3 (x) (axz) (z -. size);
  glVertex3 (x) (axz) (z +. size);
  glVertex3 (x) (axz +. size *. 2.0) (z +. size);
  glVertex3 (x) (axz +. size *. 2.0) (z -. size);
  glColor3 1.0 1.0 1.0;
  glEnd();
  error();
  (* printf "flagat9\n%!"; *)
)
;;

let displaytile i j = 
  (* i, j are indices into the array or tiles.
     (float_of_int i, float_of_int j) are the coordinates of the
     centre of tile (i, j)
  *)
  (* tilex, tiley are the indices of the tile in the texture array *)
  (
    let p = corners.(i).(j + 1) * 2 + corners.(i + 1).(j + 1) in
    let q = corners.(i).(j) * 2 + corners.(i + 1).(j) in
    tileatss p q i j;
    if flagcorner.(i).(j)
    then flagat (x_of_corner i j) (z_of_corner i j);
    (*   flagat (float_of_int i -. 0.5) (float_of_int j -. 0.5); *)
  )

let displaytiles = fun () -> (* i, j are world coordinates *)
for i = 0 to worldwidth - 1 do
  for j = 0 to worldwidth - 1 do
    displaytile i j
  done
done

let display () =
  glClear [GL_COLOR_BUFFER_BIT; GL_DEPTH_BUFFER_BIT];
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity();
  gluPerspective
     ~fovy:60.0
     ~aspect:((float_of_int !windowwidth) /. (float_of_int !windowheight))
     ~zNear:0.1
     ~zFar:(float_of_int worldwidth) (* originally 8.0 *);
  glMatrixMode GL_MODELVIEW;
  glLoadIdentity();
  glEnable GL_DEPTH_TEST;
  (* gluLookAt 0.0 0.0 2.5 0.0 0.0 0.0 0.0 1.0 0.0; *)
  if !lookdown
  then (gluLookAt ~eyeX:(!locx) ~eyeY:(altitude !locx !locy +. tall) ~eyeZ:(!locy)  
                  ~centerX:(!locx) ~centerY:(-1.0) ~centerZ:(!locy)
                  ~upX:((sin !anglex)) ~upY:0.0  ~upZ:((cos !anglex))
       )
  else (gluLookAt (!locx) (altitude !locx !locy +. tall) (!locy)
                  (!locx +. (sin !anglex) *. (cos !angley))
                      ( altitude !locx !locy +. tall +. (sin !angley))
                      (!locy +. (cos !anglex) *. (cos !angley))
                  ( -. sin !angley *. sin !anglex) (cos !angley) (-. sin !angley *. cos !anglex)
       );

glEnable GL_FOG;
(*
glFog (GL_FOG_MODE GL_LINEAR);
glFog (GL_FOG_START 1.0);
glFog (GL_FOG_END 10.0);
*)


glFog (GL_FOG_MODE GL_EXP);
glFog (GL_FOG_DENSITY 0.2);

glFog (GL_FOG_COLOR (0.5, 0.5, 1.0, 0.0));

  glEnable GL_BLEND;
  glColor3v (1.0, 1.0, 1.0);

  displaytiles();

(* the texture map, standing up like a billboard

  glBegin GL_QUADS;
  glTexCoord2 0.0 0.0;   glVertex3 (-0.25) ( 0.25) sealevel;
  glTexCoord2 0.0 1.0;   glVertex3 (-0.25) (-0.25) sealevel;
  glTexCoord2 1.0 1.0;   glVertex3 ( 0.25) (-0.25) sealevel;
  glTexCoord2 1.0 0.0;   glVertex3 ( 0.25) ( 0.25) sealevel;
  glEnd();
*)

(* Now for the map and score panel *)

(* printf "bar1%!"; *)
  error();
  glMatrixMode (GL_PROJECTION);
(*   printf "bar15%!"; *)
  error();

  glLoadIdentity();
(*   printf "bar2%!"; *)
  error();
(*  gluOrtho2D 0.0 640.0 0.0 480.0; *)
  gluOrtho2D 0.0 (float_of_int !windowwidth) 0.0 (float_of_int !windowheight); 
(*   printf "bar3%!"; *)
  error();
  glMatrixMode GL_MODELVIEW;
(*   printf "bar35%!"; *)
  error();
  glLoadIdentity();
(*   printf "bar4%!"; *)
  error();
  glDisable GL_DEPTH_TEST;
(*   printf "bar5%!"; *)
  error();
(*   printf "bar55%!"; *)

(* test squares, no longer needed 
  glBegin GL_QUADS;
(*   printf "bar6%!"; *)
(*   error(); *)
(*   printf "bar7%!"; *)
  glColor3 0.0 1.0 0.0;
(*  error(); *)
(*   printf "bar8%!"; *)
  glVertex2 1.0 1.0;
  glVertex2 40.0 1.0;
  glVertex2 40.0 40.0;
  glVertex2 1.0 40.0;
(*  error(); *)
(*   printf "bar81%!"; *)
  glEnd();
  error();
(*   printf "bar85%!"; *)

  glBegin GL_QUADS;
  glColor3 1.0 1.0 0.0;
  glVertex2 50.0 1.0;
  glVertex2 90.0 1.0;
  glVertex2 90.0 40.0;
  glVertex2 50.0 40.0;
(*  error(); *)
(*   printf "bar9%!"; *)
  glEnd();
*)

  glBegin GL_POINTS;
  let pointit i j r g b =
    (
      glColor3 r g b;
      let x = i * 2 in
      let y = ((worldwidth - j * 2)+ !windowheight - worldwidth) in

      glVertex2 (float_of_int x) (float_of_int (y));
      glVertex2 (float_of_int (x + 1)) (float_of_int y);
      glVertex2 (float_of_int x) (float_of_int (y - 1));
      glVertex2 (float_of_int (x + 1)) (float_of_int (y - 1));
    )in
  drawmap pointit;
(* -- for debugging only
  let altplot i j r g b =
    (pointit (i + 100) j r g b) in
  drawhmap altplot;
*)
  glEnd();
  error();
(*   printf "bar91%!"; *)
  glutSwapBuffers();
  error();
(*   printf "bar92%!"; *)
;;


let reshape ~width:w ~height:h =
  (* Parameters of perspective projection *)
  glViewport 0 0 w h;
  glMatrixMode GL_PROJECTION;
  glLoadIdentity();
  gluPerspective (* This is done elsewhere -- refactor ??? *)
     ~fovy:60.0 ~aspect:((float w) /. (float h))
     ~zNear:0.2
     ~zFar:(float_of_int worldwidth) (* originally 8.0 *);
  windowwidth := w; windowheight := h;
  glMatrixMode GL_MODELVIEW;
;;

let on_water x y = (* is (x, y) on water? *)
  false (* for now -- need to look up tile and check on a pixel. *)
;;

let move f r = (* f distance forward, r distance right *) 
   let newlocx = !locx +. f *. (sin !anglex) -. r *. (cos !anglex) in
   let newlocy = !locy +. f *. (cos !anglex) +. r *. (sin !anglex) in
   if newlocx > 0.0 && newlocx < (float_of_int worldwidth)
     && newlocy > 0.0 && newlocy < (float_of_int worldwidth)
     && not (on_water newlocx newlocy)
     && istraversible newlocx newlocy
   then
     ( locx := newlocx;
       locy := newlocy;
       ignore(hasflag !locx !locy)
     )
;;

let move_forward s = move s 0.0;
;;

(* active mouse motion *)
let motion ~x ~y =
  if !b_down then  (* if the left button is down *)
  begin
 (* change the rotation angle and location according to the last position
    of the mouse and the new one *)
    anglex := !anglex +. ((float_of_int (!xold - x)) /. 120.0);
    (* was: move_forward ((float_of_int (!yold - y)) *. 0.005); *)
    angley := !angley +. ((float_of_int (!yold - y)) /. 120.0);
    glutPostRedisplay();
  end;
  xold := x;  (* save mouse position *)
  yold := y;
;;

(* mouse button event *)
let mouse ~button ~state ~x ~y =
  match button, state with
  (* if we press the left button *)
  | GLUT_LEFT_BUTTON, GLUT_DOWN ->
      b_down := true;
      xold := x;  (* save mouse position *)
      yold := y;
  (* if we release the left button *)
  | GLUT_LEFT_BUTTON, GLUT_UP ->
      b_down := false;
  | _ -> ()
;;

let mvforward() = 
  ( (* printf "mvf\n"; *)
     move_forward 0.1;
     glutPostRedisplay();
  )
;;
let mvbackward() = (move (-0.05) (0.0); glutPostRedisplay());
;;
let mvleft() = (move 0.0 (-0.05); glutPostRedisplay());
;;
let mvright() = (move 0.0 (0.05); glutPostRedisplay());
;;
let turnleft() = anglex := !anglex +. 0.03; glutPostRedisplay();
;;
let turnright() = anglex := !anglex -. 0.03; glutPostRedisplay();
;;
let keyboard ~key ~x ~y =
  (* printf "key %d %c\n%!" (Char.code (key)) key; *)
  match key with
  | '\027' (* escape key *)
  | 'D' -> lookdown := not !lookdown; 
           glutPostRedisplay();
  | 'u' -> turnleft()
  | 'i' -> mvforward()
  | 'o' -> turnright()
  | 'j' -> mvleft()
  | 'k' -> mvbackward()
  | 'l' -> mvright()
  | 'q' -> turnleft()
  | 'w' -> mvforward()
  | 'e' -> turnright()
  | 'a' -> mvleft()
  | 's' -> mvbackward()
  | 'd' -> mvright()


  | _ -> (printf "key %x%c\n" (Char.code key) key)
;;

let () =
  (* texture image to load *)

  let texture, width, height, internal_format, pixel_data_format =
      Png_loader.load_img (Filename "/home/hendrik/dv/fun/wander/tiles.png")
  in
  assert_size ~width ~height;

  (* create the OpenGL window *)
  ignore(glutInit Sys.argv);
  glutInitDisplayMode [GLUT_RGBA; GLUT_DOUBLE; GLUT_DEPTH];
  glutInitWindowSize ~width:640  ~height:480;
  ignore(glutCreateWindow ~title:"wander");

  (* initialise OpenGL *)
  glClearColor ~r:0.5 ~g:0.5 ~b:1.0 ~a:0.0;
  glShadeModel GL_FLAT;
  glEnable GL_DEPTH_TEST;

  (* for textures with alpha channel *)
  glEnable GL_BLEND;
  glBlendFunc Sfactor.GL_SRC_ALPHA  Dfactor.GL_ONE_MINUS_SRC_ALPHA;
  glPolygonMode GL_FRONT_AND_BACK  GL_FILL;

  (* assumes clean models *)
  glFrontFace GL_CCW;

  glCullFace GL_FRONT_AND_BACK;  (* for alpha *)

  let texid = glGenTexture() in
  glBindTexture BindTex.GL_TEXTURE_2D texid;

  (* cleanup *)
  at_exit(fun () -> glDeleteTexture texid);

  (* Parameters for applying the textures *)
  glTexParameter TexParam.GL_TEXTURE_2D
                 (TexParam.GL_TEXTURE_MAG_FILTER  Mag.GL_NEAREST);
  glTexParameter TexParam.GL_TEXTURE_2D
                 (TexParam.GL_TEXTURE_MIN_FILTER  Min.GL_NEAREST);
  glTexImage2D TexTarget.GL_TEXTURE_2D  0  internal_format  width height
               pixel_data_format  GL_UNSIGNED_BYTE  texture;

(*  glEnable GL_TEXTURE_2D; *)

  (* Glut callback functions *)
  glutDisplayFunc ~display;
  glutReshapeFunc ~reshape;
  glutKeyboardFunc ~keyboard;
  glutMouseFunc ~mouse;
  glutMotionFunc ~motion;

  glutMainLoop();
;;
